import { Assert } from '../../Diagnostics/Assert.js';
import { Link } from '../../Nodes/Link.js';
import { NodeEvalContext } from '../../Nodes/NodeEvalContext.js';
export class SyncExecutionBlock {
    constructor(graphEvaluator, nextEval, syncEvaluationCompletedListener = undefined) {
        this.graphEvaluator = graphEvaluator;
        this.nextEval = nextEval;
        this.syncEvaluationCompletedListenerStack = [];
        this.graph = graphEvaluator.graph;
        if (syncEvaluationCompletedListener !== undefined) {
            this.syncEvaluationCompletedListenerStack.push(syncEvaluationCompletedListener);
        }
    }
    resolveInputValueFromSocket(inputSocket) {
        if (inputSocket.valueTypeName === 'flow') {
            throw new Error(`can not resolve input values for Eval input sockets: ${inputSocket.name}`);
        }
        if (inputSocket.links.length === 0) {
            if (inputSocket.value === undefined) {
                return this.graph.registry.values
                    .get(inputSocket.valueTypeName)
                    .creator();
            }
            return inputSocket.value;
        }
        if (inputSocket.links.length > 1) {
            throw new Error(`input socket has too many links: ${inputSocket.name} has ${inputSocket.links.length} links`);
        }
        const upstreamLink = inputSocket.links[0];
        const upstreamNode = this.graph.nodes[upstreamLink.nodeId];
        const upstreamOutputSocket = upstreamNode.getOutputSocket(upstreamLink.socketName);
        if (upstreamNode.flow) {
            inputSocket.value = upstreamOutputSocket.value;
            return upstreamOutputSocket.value;
        }
        upstreamNode.inputSockets.forEach((upstreamInputSocket) => {
            this.resolveInputValueFromSocket(upstreamInputSocket);
        });
        const context = new NodeEvalContext(this, upstreamNode);
        context.evalImmediate();
        inputSocket.value = upstreamOutputSocket.value;
        return inputSocket.value;
    }
    commit(outputFlowSocket, syncEvaluationCompletedListener = undefined) {
        Assert.mustBeTrue(this.nextEval === null);
        const node = this.graph.nodes[outputFlowSocket.nodeId];
        const outputSocket = node.getOutputSocket(outputFlowSocket.socketName);
        if (outputSocket.links.length > 1) {
            throw new Error('invalid for an output flow socket to have multiple downstream links:' +
                `${node.typeName}.${outputSocket.name} has ${outputSocket.links.length} downlinks`);
        }
        if (outputSocket.links.length === 1) {
            const link = outputSocket.links[0];
            if (link === undefined) {
                throw new Error('link must be defined');
            }
            this.nextEval = link;
        }
        if (syncEvaluationCompletedListener !== undefined) {
            this.syncEvaluationCompletedListenerStack.push(syncEvaluationCompletedListener);
        }
    }
    executeStep() {
        const link = this.nextEval;
        this.nextEval = null;
        if (link === null) {
            if (this.syncEvaluationCompletedListenerStack.length === 0) {
                return false;
            }
            const awaitingCallback = this.syncEvaluationCompletedListenerStack.pop();
            if (awaitingCallback === undefined) {
                throw new Error('awaitingCallback is empty');
            }
            awaitingCallback();
            return true;
        }
        const node = this.graph.nodes[link.nodeId];
        node.inputSockets.forEach((inputSocket) => {
            if (inputSocket.valueTypeName !== 'flow') {
                this.resolveInputValueFromSocket(inputSocket);
            }
            else {
                inputSocket.value = inputSocket.name === link.socketName;
            }
        });
        const context = new NodeEvalContext(this, node);
        context.evalFlow();
        if (context.numCommits === 0 && !context.asyncPending) {
            let numFlowOutputs = 0;
            node.outputSockets.forEach((outputSocket) => {
                if (outputSocket.valueTypeName === 'flow') {
                    numFlowOutputs++;
                }
            });
            if (numFlowOutputs !== 1) {
                throw new Error(`can not use auto-commit if there are multiple flow outputs, number of outputs is ${numFlowOutputs} on ${node.typeName}`);
            }
            node.outputSockets.forEach((outputSocket) => {
                if (outputSocket.valueTypeName === 'flow') {
                    this.commit(new Link(link.nodeId, outputSocket.name));
                }
            });
        }
        return true;
    }
}
//# sourceMappingURL=SyncExecutionBlock.js.map