import { Assert } from '../../Diagnostics/Assert.js';
import { EventEmitter } from '../../Events/EventEmitter.js';
import { AsyncNode } from '../../Nodes/AsyncNode.js';
import { EventNode } from '../../Nodes/EventNode.js';
import { sleep } from '../../sleep.js';
import { Fiber } from './Fiber.js';
export class Engine {
    constructor(graph) {
        this.graph = graph;
        this.fiberQueue = [];
        this.asyncNodes = [];
        this.eventNodes = [];
        this.onNodeExecution = new EventEmitter();
        this.executionSteps = 0;
        Object.values(graph.nodes).forEach((node) => {
            if (node instanceof EventNode) {
                this.eventNodes.push(node);
            }
        });
        this.eventNodes.forEach((eventNode) => eventNode.init(this));
    }
    dispose() {
        this.asyncNodes.forEach((asyncNode) => asyncNode.dispose());
        this.eventNodes.forEach((eventNode) => eventNode.dispose(this));
    }
    commitToNewFiber(node, outputFlowSocketName, fiberCompletedListener = undefined) {
        Assert.mustBeTrue(node instanceof EventNode || node instanceof AsyncNode);
        const outputSocket = node.outputSockets.find((socket) => socket.name === outputFlowSocketName);
        if (outputSocket === undefined) {
            throw new Error(`no socket with the name ${outputFlowSocketName}`);
        }
        if (outputSocket.links.length > 1) {
            throw new Error('invalid for an output flow socket to have multiple downstream links:' +
                `${node.description.typeName}.${outputSocket.name} has ${outputSocket.links.length} downlinks`);
        }
        if (outputSocket.links.length === 1) {
            const fiber = new Fiber(this, outputSocket.links[0], fiberCompletedListener);
            this.fiberQueue.push(fiber);
        }
    }
    executeAllSync(limitInSeconds = 100, limitInSteps = 100000000) {
        const startDateTime = Date.now();
        let elapsedSeconds = 0;
        let elapsedSteps = 0;
        while (elapsedSteps < limitInSteps &&
            elapsedSeconds < limitInSeconds &&
            this.fiberQueue.length > 0) {
            const currentFiber = this.fiberQueue[0];
            const startingFiberExecutionSteps = currentFiber.executionSteps;
            currentFiber.executeStep();
            elapsedSteps += currentFiber.executionSteps - startingFiberExecutionSteps;
            if (currentFiber.isCompleted()) {
                this.fiberQueue.shift();
            }
            elapsedSeconds = (Date.now() - startDateTime) * 0.001;
        }
        this.executionSteps += elapsedSteps;
        return elapsedSteps;
    }
    async executeAllAsync(limitInSeconds = 100, limitInSteps = 100000000) {
        const startDateTime = Date.now();
        let elapsedSteps = 0;
        let elapsedTime = 0;
        let iterations = 0;
        do {
            if (iterations > 0) {
                await sleep(0);
            }
            elapsedSteps += this.executeAllSync(limitInSeconds - elapsedTime, limitInSteps - elapsedSteps);
            elapsedTime = (Date.now() - startDateTime) * 0.001;
            iterations += 1;
        } while ((this.asyncNodes.length > 0 || this.fiberQueue.length > 0) &&
            elapsedTime < limitInSeconds &&
            elapsedSteps < limitInSteps);
        return elapsedSteps;
    }
}
//# sourceMappingURL=Engine.js.map